package Search::OpenFTS::Crawler;
use strict;

=head1 NAME

 Search::OpenFTS::Crawler - Base Crawler class for OpenFTS

=cut

=head1 PREREQUISITIES

 DBI, DBD::Pg, URI, libwww-perl, Digest::MD5 - available from CPAN
 Search::OpenFTS::Index 

=cut

=head1 SYNOPSIS

 Simple example to crawl and index 10 pages
 (see crawler.pl for more sophisticated crawler).

 use strict;
 # provide path to your custom dictionaries
 use lib qw (/u/megera/app/fts/test-suite);
 use Search::OpenFTS::Crawler;
 my $parent = 'http://www.foo.com/';
 my $crawler = Search::OpenFTS::Crawler->new(dbname=>'www',
                                             parent=>$parent,
                                             maxpages=>10,
                                             delay => 0.01,
                                             op=>'add',
                                             );
                                            
 my ($page, $errstr);
 my $npage = 0;
 my $ua = $crawler->{ua};
 my $uri = $parent;
 while ( defined $uri && $npage < $crawler->{maxpages} ) {
   $page = $crawler->fetch_page ( $ua, $uri );
   if ( $page->{success} && $page->{content_type} =~ /^text\// ) {
     $errstr = $crawler->index($crawler->{tid},$page);
     if ( $errstr ) {
       warn "$errstr:$uri\n";
     } else {
       $npage++;
     }
     $crawler->getlinks($page->{base}, $page->{content});
   } else {
     warn "Return code:", $page->{code}, ",$uri\n";
   }
   $crawler->mark_url($uri, ( $page->{success} ) ? $page->{code} : -$page->{code});
   $uri = $crawler->next_url();
 }

 $crawler->dump_links(\*STDOUT);
 $crawler->finish();


=cut

=head1 DESCRIPTION

 my $crawler = Search::OpenFTS::Crawler->new (attrib1 => $attrib1,
                                              ...................,
                                             );

Constructor accepts following attributes:

=over

=item dbname

Database name. This is B<obligatory> attribute.
Database handler could be accessed as $crawler->{dbh}

=item maxpages

Maximum number of pages to be indexed. Pages rejected by user defined rules,
robots.txt or unsusccessfuly fetched don't counted. Default is 10 pages.
You may changed it after creating crawler object: $crawler-{maxpages} = 100;

=item parent

URL of parent page. This is B<obligatory> attribute.
Used by B<check_user_rules> method for rejecting offsite links. 
Accessed as $crawler->{parent}.


=item debug

Used for various debug printings. Accepts {0,1,2} values.
The higher debug level, the more printings.

=item op

Controls crawling and indexing policies. Accepts {'add', 'reindex', 'noindex'} values. 

=over

=item * 'undef' ( default )

Used only for the first time you index *specific* collection. 
Remember, OpenFTS supports up to 27
different collections in one database. Collection is specified by
B<prefix> attribute (see below). Accessed as $crawler->{op}.

=item * 'add'

Used to continue crawling. Previously indexed pages will not be touched !
Could be used in conjunction with B<load_links> method,
which loads information about not-indexed pages from previous run
(see B<dump_links> method). 

=item * 'reindex'

Used to reindex already indexed collection. No new pages will be
crawled !

=item * 'noindex'

Used for gathering links. Database should not be touched. Example:

$errstr = $crawler->index($crawler->{tid},$page) if $crawler->{op} ne 'noindex');

=back

=item patterns

Reference to array with users defined patterns (perl regexps). Used in
B<check_user_rules> method to exclude pages. Example:

 my @exclude_patterns = qw (  /db/apod.html\?d=.*&page=\d* \.bad_ext$ );
 my $crawler = Search::OpenFTS::Crawler (  patterns => \@exclude_patterns, ....);

You may test your exclusion rules, see documentation to MyCrawler.pm

=item delay

Delay time between requests to the same server (in minutes). 
Default is 0.5. See B<LWP::RobotUA> for more information.

=item prefix

a single latin character {a..z}. Used to specify collection to work with.
Default is ''.  OpenFTS works with auxiliary tables required for 
storing internal information. Also, collection itself is stored in 
prefixed table. Collections are fully independent and created in
initialization stage (see init-crawler.pl for more information).

=back

=cut

use DBI;
use URI::URL qw(url);
use URI;
use LWP::UserAgent;
#use LWP::RobotUA;
#require WWW::RobotRules;
use Search::OpenFTS::Index;
use Digest::MD5 qw(md5_base64);

use vars qw($VERSION);
local $URI::URL::ABS_REMOTE_LEADING_DOTS = 1;

$VERSION = '0.4.1';

sub new {
  my ($self, %opts) = @_;
  my $class = ref($self) || $self;
  $self = {} unless ref $self;
  bless $self, $class;

  my $tid = 0;
  $self->{op} = $opts{op};
  $self->{VERSION} = $VERSION;
  $self->{PREFIX} = $opts{prefix};

  if ( $self->{op} ne 'noindex' ) {
    my $dbname = $opts{dbname} || die "No dbname specified !\n";
    my $dbh = DBI->connect('DBI:Pg:dbname='.$dbname,"","", { AutoCommit => 0, PrintError =>0 }) || die "Couldn't connect to db: $dbname !\n";
    $self->{dbh} = $dbh;
    $self->{idx} = Search::OpenFTS::Index->new($dbh, $self->{PREFIX});
    $self->{TABLE} = $self->{idx}->{TABLE};
    $self->{TXTID} = $self->{idx}->{TXTID};
    $self->{PREFIX} = $self->{idx}->{PREFIX};
# prepare sql patterns
    $self->sql_patterns();
# update sql patterns - change table, txtid names
    foreach ( keys %{ $self->{sql} } ) {
       $self->{sql}->{$_} =~ s/#txt#/$self->{TABLE}/;
       $self->{sql}->{$_} =~ s/#tid#/$self->{TXTID}/ if ! /^insert/;
       $self->{sth}->{$_} = $dbh->prepare ( $self->{sql}->{$_} ) or $self->mydie($dbh->errstr); 
    }
    $tid=$dbh->selectrow_array($self->{sql}->{max_tid}) || 0;
  }

    $self->{tid} = 0;
    $self->{tid} = ( $tid ) ? $tid : 0 if ( $self->{op} eq 'add' );
    $self->{maxpages} = $opts{maxpages};
    $self->{maxpages} = ( $tid ) ? $tid : 0 if ( $self->{op} eq 'reindex' );

    $self->{AGENT} = 'OpenFTS Crawler/'.$VERSION.' ';
# create UA object
    my $ua = new LWP::UserAgent; 
    $ua->max_size(100000); # restrict page size by 100Kb
    $ua->agent($self->{AGENT});
#    my $ua = new LWP::RobotUA $self->{AGENT}, 'your@email';
#    $ua->delay( $opts{delay} || 0.5);
    $self->{ua} = $ua;

    if ( $self->{op} ne 'reindex' && $opts{parent} ) {
# check parent page and fix URL for possible redirection
      my $parent = $opts{parent};
      my $request  = HTTP::Request->new( HEAD => $parent );
      my $response = $self->{ua}->request($request);
      warn "success:", $response->is_success,"\n";
      if ( $response->is_success ) {
        $parent = ${ $response->request->url };
      } else {
#        $self->mydie ("Problem with fetching of parent page \'$parent\' !\n");
      }
      $parent =~ s/#.*$//;    # remove fragments
      $parent =~ s/\?.*$//;   # remove cgi part
      $parent = url($parent)->canonical;
      $self->{parent} = $parent;
# get robots rules  
#      $self->robots($self->{parent});
    }

# user patterns, compile rules regexp for performance
  $self->{patterns} = $opts{patterns};
  my $op; my @expr = ();
  foreach ( 0..$#{$self->{patterns}} ) {
    if ( $self->{patterns}[$_] =~ /^!(.*)/ ) {
      $op = '!';
      $self->{patterns}[$_] = $1;
   } else {
      $op = '=';
   }
   push @expr, "\$url ". $op ."~ m/\$self->{patterns}[$_]/o";
  }
  my $expr = join('||', @expr);
#  my $expr = join('||', map {"\$url =~ m/\$self->{patterns}[$_]/o"} 0..$#{$self->{patterns}} );
  $expr ||= 0; # don't match anything if no patterns specified
  $self->{rules} = eval "sub { my \$url = shift; $expr }";
#warn $self->{patterns}[0],"\n";
#warn "rules:",$expr,"\n";
  die "$@" if $@; 

  $self->{DEBUG} = $opts{debug};
  $self->{DEBUG} ||= 0;

  return $self;
}

=head1 METHODS

=over

=item B<index>($tid,$page)  

Index/reindex ( $self->{op} eq 'reindex' )  page (see B<fetch_page> for
description of $page . Wrapper around delete(), insert() and _index() methods.

Returns error string if something fails (all changes will be rollbacked).

=cut

sub index {
  my ($self, $tid, $page) = @_;
  my $errstr = '';

  if ( $self->{op} eq 'reindex') { # delete and insert in transaction
    $tid = $self->{linkstid}->{$page->{uri}};
    $errstr = $self->delete($tid,commit=>0); # no commit !
    return $errstr if $errstr;
  } else { # need new tid
    $tid++;
  }

  warn "Indexing:$tid,", $page->{uri},"\n";

  $errstr = $self->insert($tid,$page,commit=>0);
  $errstr ||= $self->_index($tid, $page->{content},$page->{title});

# everything is ok, store $tid
  $self->{tid} = $tid if ( !$errstr ); 

  return $errstr;
}

=item B<_index>($tid, $body, $title) 

Internal method for indexing. This method inherits index() methods from 
Search::OpenFTS::Index.

B<Warning>: the method commits all changes made before its invoking !
Returns error string if indexing fails ( all changes are rollbacked !)

=cut

sub _index{
  my ( $self, $tid, $body, $title ) = @_;
  my $errstr = "";
  my $dbh = $self->{dbh};
  my $rc = $self->{idx}->index($tid, $body, $title);
  if ( !$rc || $rc eq '0E0' ) {
       $errstr =  $dbh->errstr;
       $dbh->rollback;
  } else { # finally commit all changes !
       $dbh->commit;
  }
  return $errstr;
}

=item B<delete>($tid, $page, commit=>0|1) 

delete and deindex page. 

This operation is consist of
2 steps - delete page record from table and delete according
entries from index tables. Commit changes if commit = 1;

Returns error string if deindexing fails

=cut

sub delete {
  my ( $self, $tid, %opt ) = @_;
  my $errstr = '';
  my $dbh = $self->{dbh};
  if ( !( $self->{sth}->{delete}->execute($tid) &&  $self->{idx}->delete($tid) ) ) 
  {
     $errstr = $dbh->errstr;
     $dbh->rollback;
  } elsif ( $opt{commit} ) { # commit changes
     $dbh->commit;
  }
  return $errstr;
}

=item B<sql_patterns>  

Defines sql patterns used for manipulation with database.

=over

=item * 

insert_row - insert row into table

=item *

max_tid    - select maximum value of text id

=item *

links      - select all pairs (path,tid)

=item *

url        - select (path,tid) for given path

=item *

md5        - select text id for given md5 checksum

=item *

delete     - delete row for give text id

=back

See also B<insert> for description of default table schema and obligatory 
fields.

B<NOTICE>: This method should be overloaded to reflect user specific scheme.
(see B<CUSTOMIZATION> section )

=cut

sub sql_patterns {
  my ( $self ) = shift;

  $self->{sql}->{insert_row} = "insert into #txt# ( tid, path, body, di, dlm, de, md5 ) values ( ? , ?, ?, ?, ?, ?, ? );";
  $self->{sql}->{max_tid} = "select max(#tid#) from #txt#";
  $self->{sql}->{links} = "select path,#tid# from #txt#";
  $self->{sql}->{url} = "select path,#tid# from #txt# where path = ?";
  $self->{sql}->{md5} = "select #tid# from #txt# where md5 = ?";
  $self->{sql}->{delete} = "delete from #txt# where #tid# = ?";

}


=item B<insert>($tid, $page, commit=>0|1) 

Insert new page into table

Commit changes if commit => 1 . Returns error string if insert fails

Default schema is: 

=over

=item *

tid  - int4 - text identificator ( B<obligatory> )

=item *

path - varchar - url ( B<obligatory> )

=item *

body - varchar - content of page ( B<obligatory> )

=item *

di   - int4 - timestamp of indexing ( B<obligatory> )

=item *

dlm  - int4 - last modified || di  ( B<obligatory> 

=item *

de   - int4 - expiration date   || di   ( B<obligatory> )

=item *

md5  - char(22) - md5 checksum      ( B<obligatory> )

Also, there is obligatory field B<fts_index> of type txtidx used by indexing 
procedure. 

di, dlm, de - in seconds since 1970 

=back

B<NOTICE>: This method should be overloaded to reflect table structure if 
default scheme is not what you need. (see B<CUSTOMIZATION> section )

=cut

sub insert {
  my ( $self, $tid, $page, %opt) = @_;
  my $errstr = '';
  my $dbh = $self->{dbh};
  $self->{sth}->{insert_row}->execute(
                     $tid,
                     $page->{uri},
                     $page->{content},
                     time(), 
                     $page->{last_modified},
                     $page->{expires},
                     $page->{md5});
  $errstr = $dbh->errstr;
  $dbh->rollback if ( $errstr );
  $dbh->commit if ( $opt{commit} );  
  return  $errstr;
}


=item B<url_exists>($url) 

returns $tid (text identificator) if $url is already exists and 
'undef' if $url is not found.

=cut

sub url_exists {
  my ( $self, $url ) = @_;
  $self->{sth}->{url}->execute($url);
  return undef if $self->{dbh}->errstr;
  my ( $url_dummy, $tid ) = $self->{sth}->{url}->fetchrow_array;
  return undef if $self->{dbh}->errstr;
  return $tid;
}


=item B<md5_exists>($md5) 

returns $tid (text identificator) if checksum $md5 is already exists and 'undef' 
if $md5 is not found.

=cut

sub md5_exists {
  my ( $self, $md5 ) = @_;
  $self->{sth}->{md5}->execute( $md5 );
  return undef if $self->{dbh}->errstr;
  my $tid = $self->{sth}->{md5}->fetchrow_array;

  warn "Not modified:tid=$tid\n" if ( $self->{DEBUG} >= 1 && $tid );

  return undef if $self->{dbh}->errstr;
  return $tid;
}

=item B<getlinks>($base, $body)  

parse $body and extract href's

urls converted to absolute canonical form using $base url and cleanuped.
After checking for robots ( B<check_robot_rules> ) and user defined 
( B<check_user_rules> ) rules they are stored as:
 
$self->{url}->{$url} = 1;


=cut

# simple, fast parser
sub getlinks {
  my ( $self, $base, $body ) = @_;
  my $url;
#  while ( $body =~ m/href\s*=\s*["'](.*?)['"]/gisx ) {
  while ( $body =~ m/href\s*=\s*["']?([^"'>]*)['"]?/gisx ) {
    $url = url($+, $base)->abs->canonical;
print STDERR "extract URL:$url\n" if  $self->{DEBUG} == 2;
    next if ( $url !~ /^http:/i );
# get rid fragments
    $url =~ s/#.*$//;
    $url =~ s/&amp;/&/g;
    $url =~ s/%3F/?/g;
# fix possible CR's in url - workaround for url hyphenation
    $url =~ s/(%0A)//g;
print STDERR "add URL:$url\n" if $self->{DEBUG} == 2;
    $self->add_link($url);
  }
}

=item B<getlinks_db>()  

get links from database

B<Warning>: It's assumed that all links in database were already checked
This method could be used for reindexing.

$self->{linkstid}->{$url} is used to store $tid ( text identificator )

$self->{expire}->{$url} stores expiration time

=cut


# get links from database for reindexing, no checks !
sub getlinks_db {
  my $self = shift;
  foreach  ( @{ $self->{dbh}->selectall_arrayref($self->{sth}->{links})  } ) {
     my $url = $_->[0];
     $self->add_link($url);
     $self->{linkstid}->{$url} = $_->[1]; # save tid
     $self->{expire}->{$url} = $_->[2]; # save expiration date
  }
}

=item B<add_link>($uri)

Add $uri to queue. Checks for robot and user exclusion rules.
Check if $uri is already in queue

=cut

sub add_link {
  my ( $self, $uri) = @_;
#  return undef if !$self->check_robot_rules( $uri );
  return undef if !$self->check_user_rules ( $uri );
# keep previous value if exists
  return undef if exists $self->{links}->{$uri};
  $self->{links}->{$uri} = 1;
}

=item B>delete_link>($uri)

Delete $uri from queue

=cut

sub delete_link {
  my ( $self, $uri) = @_;
  delete $self->{links}->{$uri};
}


=item B<mark_url>($uri, $code)  

mark $uri as indexed.

 $code > 0 - if page was indexed successfully
 $code < 0 - if there were problem with fetching or indexing
 $code should not be '1', because this value is used for internal
 purposes, see B<next_url>.

=cut

sub mark_url {
  my ( $self, $uri, $code ) = @_;
  $self->{links}->{$uri} = $code if ( $uri );
}


=item B<next_url>($urlprefix) 

returns next not-indexed $uri or 'undef' if queue is empty.
See also B<mark_url>

=cut

sub next_url {
  my ( $self, $urlprefix ) = @_;
  foreach my $uri ( grep ( /^$urlprefix/, keys %{$self->{links}} ) ) {
    return $uri if ( $self->{links}->{$uri} == 1 );
  }
  return undef; # queue is empty
}

=item B<robots>($uri) 

fetch /robots.txt and stores robots exclusions rules.
Returns 'undef' if /robots.txt doesn't exists.

=cut

sub robots {
  my ($self, $uri ) = @_;
  my $robotsrules = new WWW::RobotRules($self->{AGENT});
  my $ua = $self->{ua};
  my $u = URI->new($uri);
  my $url = $u->scheme.'://'.$u->host.'/robots.txt';
  my $request  = HTTP::Request->new( GET => $url);
  my $response = $ua->request($request);
  return undef if !$response->is_success;
  $robotsrules->parse($url, $response->content);
  $self->{robotrules} = $robotsrules;
  return 1;
}

=item B<check_robot_rules>($uri,$reason) 

check robot rules

Returns FALSE if $uri should be excluded, store message in $reason
Returns TRUE if $url is ok.

 Usage: 
 print "$reason\n" if ( !check_robot_rules($url, \$reason) );
 print "Forbidden\n" if ( !check_robot_rules($url) );

=cut

sub check_robot_rules {
  my ( $self, $url, $reason ) = @_;
  if ( defined $self->{robotrules} && !$self->{robotrules}->allowed($url) ) {
    $$reason = "Forbidden by robots.txt";
    return 0;
  }
  return 1;
}

=item B<check_user_rules>($uri,$reason) 

check user defined rules

Returns FALSE if $uri doesn't satisfy the user defined rules.
This method could be overloaded by user.

 Default rules:
  
 1. Ignore $uri if it doesn't satify the users exclusion patterns
 2. Ignore $uri if it's not under parent page
 3. Ignore some extensions (jpg|gif|pdf|ps|css|tif|zip|arj|tar|gz|lha)
 4. Ignore Apache auto-indexing tricks

 Usage: 
 print "$reason\n" if ( !check_user_rules($url, \$reason) );
 print "Forbidden\n" if ( !check_user_rules($url) );

=cut

sub check_user_rules {
  my ( $self, $url, $reason ) = @_;

# check users patterns
  if ( ref($self->{rules}) eq 'CODE' && &{ $self->{rules}} ($url) ) {
       $$reason = "Forbidden by user rules" if ref $reason eq 'SCALAR';
       return 0;
  }

# check if link is under parent url, be careful, compiled regexp !!!
  if ( $self->{parent} && $url !~ /^\Q$self->{parent}/o ) {
      $$reason = "Not under parent" if ref $reason eq 'SCALAR';
      return 0;
  }

# add your rules !
  if ( $url =~ /\.(jpg|gif|pdf|ps|css|tif|zip|arj|tar|gz|lha)$/o ) {
      $$reason = "Bad extension:.$1" if ref $reason eq 'SCALAR';
      return 0;
  }
  if ( $url =~ /\/\?(N=A|M=A|S=A|D=A|N=D|M=D|S=D|D=D)$/o ) {
      $$reason = "Ignore Apache auto-indexing:?$1" if ref $reason eq 'SCALAR';
      return 0;
  }

  
  return 1;
}

=item B<fetch_page>($ua, $uri) 

fetch $uri and returns $page object. $ua - is a reference to LWP::RobotUA or
LWP::UserAgent object.

 Usage:

 $ua = new LWP::RobotUA "OpenFTS Crawler/0.3", 'your@email';
 $page = $self->fetch_page($ua,$uri);

 
 $page->{content_length} - content length
 $page->{content_type}   - content type
 $page->{success}        - 1 is request was successfull
 $page->{last_modified}  - last modification date
 $page->{expires}        - expiration date
 $page->{code}           - return code

 Also, if $page->{success} && $page->{content_type} =~ /^text/
 
 $page->{uri}            - uri, could be changed because of redirection
 $page->{base}           - base url of page

 If $page->{uri} satisfy the robots and user defined rules

 $page->{content}        - content of page
 $page->{title}          - title
 $page->{md5}            - MD5 checksum

 See also B<insert> method. 

=cut

sub fetch_page {
  my ( $self, $ua, $uri ) = @_;

#  my $ua = $self->{ua};

# get HEADER
  my $request  = HTTP::Request->new( HEAD => $uri );
  my $response = $ua->request($request);

  warn $uri,":",$response->as_string,"\n" if ( $self->{DEBUG} == 2 );
  my $reason = "";
  my $page = {};
  $page->{content_length} = $response->content_length;
  $page->{content_type} = $response->content_type;
  $page->{success} = $response->is_success;
  $page->{last_modified} = $response->last_modified;
  $page->{last_modified} ||= time(); # current time
  $page->{expires} = $response->expires;
  $page->{expires} ||= time(); # current time
  $page->{code} = $response->code;

  return $page if $response->content_type !~ /^text\// ||
                  !$response->is_success;

  $page->{uri} = ${ $response->request->url };
  $page->{base} = $response->base;
# $uri is probably changed due to redirection ! Need check against rules
  if ( $uri ne $page->{uri} ) {  
    if ( !$self->check_robot_rules( $page->{uri}, \$reason ) ||
         !$self->check_user_rules ( $page->{uri}, \$reason )
       ) 
    {  
      warn "Redirection to location forbidden by rule:$reason:",$page->{uri},"\n";
      delete $self->{links}->{$uri}; # delete original URI from queue
      $page->{success} = 0;
      return $page;
    }
  } 

# get PAGE
  $request  = HTTP::Request->new( GET => $uri );
  $response = $ua->request($request);

  $page->{content} = $response->content;
  $page->{title} = $response->title;
  $page->{md5} = md5_base64($page->{content});
# sometimes 'content_length' doesn't returns. Calculate it myself
  $page->{content_length} ||= length($page->{content});

  return $page;
}

=item B<dump_links>($fh) 

dump B<not-indexed> links to filehandle

 Usage:

  my $file = '/path/to/file/with/links';
  if ( open(LINKS,">$file") ) {
    $crawler->dump_links(\*LINKS);
    close(LINKS);   
  } else {
    warn "Can't open $file for writing !\n";
  }

=cut

sub dump_links {
  my ($self, $fh ) = @_;
  if ( ref $fh eq 'GLOB' ) {
    map { print $fh "$self->{links}->{$_}:$_\n" if $self->{links}->{$_} <=1; } 
          keys %{$self->{links}};
  }
}

=item B<dump_links_all>($fh)  

dump B<all> links to filehandle

 Usage:

  my $file = '/path/to/file/with/links';
  if ( open(LINKS,">$file") ) {
    $crawler->dump_links_all(\*LINKS);
    close(LINKS);   
  } else {
    warn "Can't open $file for writing !\n";
  }

=cut

sub dump_links_all {
  my ($self, $fh ) = @_;
  if ( ref $fh eq 'GLOB' ) {
    map { print $fh "$self->{links}->{$_}:$_\n"; } keys %{$self->{links}};
  }
}

=item B<load_links>($fh) 

load B<not-indexed> links from filehandle

Method expects $code:$url pairs, saved in previous run of crawler
using --dump option.
Returns number of links loaded or 'undef' if method fails

 Usage:

  my $file = '/path/to/file/with/links';
  if ( open(LINKS, "<$file") ) {
    $crawler->load_links(\*LINKS);
    close(LINKS);
  } else {
    warn "Unable to open $file !\n";
  }


=cut

# load links from filehandle ( $code:$url )
sub load_links {
 my ($self, $fh ) = @_;
 if ( ref $fh eq 'GLOB' ) {
    my $i = 0;
    while (<$fh>) {
      chomp;
      my ( $code, $url ) = split (/:/, $_,2);
      $url =~ ~s/^\s*(.*?)\s*$/$1/;
      if ( $code == 1 ) {
         $self->add_link($url);
         $i++;
      }
    }
    return $i;
 } else {
    return undef;
 } 
} 

=item B<finish> 

disconnect from database and flush all active statements handles

=back

=cut 

sub finish {
  my ( $self ) = shift;
  return 1 if ( !$self->{dbh} );
  map { $self->{sth}->{$_}->finish(); } keys %{ $self->{sth} };
  $self->{dbh}->disconnect();
}


=item B<mydie>($errstr)

flush all active statement handles and die. See also finish().

=cut

sub mydie {
  my ( $self, $errstr ) = @_;
  $self->finish();
  die "$errstr\n";
}

1;


=head1 CUSTOMIZATION

Package Search::OpenFTS::Crawler provides base methods to build
crawlers and designed to work with OpenFTS full text search engine.

The best way is to write your own perl module which inherits methods from
Search::OpenFTS::Crawler and overloads some methods to suit specific
requirements. See crawler.pl for example of using custom module
( MyCrawler.pm ) in example directory.

There are three methods which are likely required to be overloaded:

=item B<check_user_rules>

Adds your specific exclusion rules to default one.

=item B<sql_patterns>

Provides sql patterns which reflects specific table schema used for
storing indexed pages. 

=item B<insert>

Provide method to insert row into table. It should be consistent with
sql pattern 'insert_row' defined in  B<sql_patterns>.


Originally, all tables for specific collection are created by 
init-crawler.pl. Typical session is:

=item 1. Modify init-crawler.pl - change table name in 

my $TABLE = 'pages';

and schema in 

  $dbi->do("create table ${TPREFIX}$TABLE (
        tid int not null primary key,
        path varchar not null unique,
        body varchar,
        title varchar,
        di int4,
        dlm int4,
        de int4,
        md5 char(22),
        fts_index txtidx
  );") || die;


Notice, all fields are B<obligatory>, you are free to change their names,
but not meanings ! Also, you may add some additional fields.
Check if your changes are consistent with 

  my $idx=Search::OpenFTS::Index->init(......);
 
Run: init-crawler.pl www:a

B<NOTICE>: Database 'www' should be exists and sql functions (from tsearch.sql 
and  openfts.sql) should be imported ! Read installation instructions and
README files in examples directory.

This will creates OpenFTS auxilliary tables and main table for storing
indexed pages ( 'a' is a prefix used for the collection - all table names will
be  ). Also, all these tables will be granted to PUBLIC (select only).
Read B<fix_permissions> method if you want to grant permissions to specific
user ( usually httpd user ).

=item 2. Customize methods

Modify MyCrawler.pm ( methods B<sql_patterns> and B<insert> )
to reflect changes in init-crawler.pl (if any).

=item 3. Actually crawling.

crawler.pl --db=www:a --maxpages=100 --dump=LINKS http://www.yoursite.com/

Inspect file LINKS which will contains URLs of all notindexed pages (if any).
If you wish to continue crawling:

crawler.pl --db=www:a --dump=LINKS --load=LINKS --op=add http://www.yoursite.com/

B<NOTICE>: crawler will try to index ALL the pages from LINKS file.

I would recomment inspect LINKS file to find 'suspicious' links you
dont' want to be indexed and define exclude patterns in crawler.pl

=item 4. Searching. 

After crawling it's possible to search using
search.pl from command line:

   search.pl -p www:a -h5 your words here

and search.cgi ( you need to change database name and prefix )

=cut


=head1 AUTHOR

 Oleg Bartunov (oleg@sai.msu.su)

=cut

=head1 SEE ALSO

 Search::OpenFTS
 
 init-crawler.pl,crawler.pl,MyCrawler.pm in examples

=cut
