package MyCrawler;

=head1 NAME

MyCrawler - Package for building web crawlers 

 Inherits classes from Search::OpenFTS::Crawler
 Overloads methods check_user_rules, sql_patterns, insert

=cut

=head1 SYNOPSIS

 # Example: test exclude rules
 use strict;
 use MyCrawler;
 # place your regexp here
 my @exclude_patterns = qw ( 
                    \.gif$ \.jpg$ \.gz$
                    /db/apod.html\?d=.*&page=\d*
                  );

 my $crawler = MyCrawler->new(
                              op       =>'noindex',
                              patterns =>\@exclude_patterns,
                             );
 my $reason;
 while (<>) {
   chomp;
   print "Yuk:$reason, $_\n" if !$crawler->check_user_rules($_, \$reason);
 }

=cut

use strict;
use Digest::MD5 qw(md5_base64);
use Search::OpenFTS::Crawler;
use vars qw(@ISA);
@ISA = qw(Search::OpenFTS::Crawler);

=head1 METHODS

=cut

=item B<check_user_rules>($url,$reason) - check user defined rules

 Returns FALSE if $url doesn't satisfy the user defined rules.
 Overloads SUPER::check_user_rules method.

 Usage: 
 print "$reason\n" if ( !check_user_rules($url, \$reason) );
 print "Forbidden\n" if ( !check_user_rules($url) );

=cut

sub check_user_rules {
  my ( $self, $url, $reason ) = @_;

# Add you rules here

# check method from SUPER class 
  return $self->SUPER::check_user_rules($url, $reason);
  
# accept $url
  return 1;
}

=item B<sql_patterns>() - sql patterns

 This method overloads SUPER::sql_patterns() to reflect user specific scheme.

=over 8
 
=item  B<insert_row> - insert row into table

=item  B<max_tid>    - select maximum value of text id

=item  B<links>      - select all pairs path,tid

=item  B<url>        - select path,tid for given path

=item  B<md5>        - select text id for given md5 checksum

=item  B<delete>     - delete row for give text id

=back

#tid# and #txt# are patterns for text identificator and table name.
Don't change them ! They are defined at initialization ( init-crawler.pl )

See also insert().

=cut

sub sql_patterns {
  my ( $self ) = shift;
  my $sql = \%{ $self->{sql} };

  $sql->{insert_row} = "insert into #txt# ( tid, path, body, title, di, dlm, de, md5 ) values ( ? , ?, ?, ?, ?, ?, ?, ? );";
  $sql->{max_tid} = "select max(#tid#) from #txt#";
  $sql->{links} = "select path,#tid#,dlm from #txt#";
#  $sql->{links} = "select path,#tid#,de from #txt#";
  $sql->{url} = "select path,#tid# from #txt# where path = ?";
  $sql->{md5} = "select #tid# from #txt# where md5 = ?";
  $sql->{delete} = "delete from #txt# where #tid# = ?";

}

=item B<insert>($tid, $page, commit=>0|1) - insert new page

 This method overloads SUPER::insert to reflect table structure

 Commit changes if commit => 1
 Returns error string if insert fails

=cut

sub insert {
  my ( $self, $tid, $page, %opt) = @_;
  my $errstr = '';
  my $dbh = $self->{dbh};
  $self->{sth}->{insert_row}->execute(
                     $tid,
                     $page->{uri},
                     $page->{content},
                     $page->{title},
                     time(),
                     $page->{last_modified}, 
                     $page->{expires},
                     $page->{md5});
  $errstr = $dbh->errstr;
  $dbh->rollback if ( $errstr );
  $dbh->commit if ( $opt{commit} );  
  return  $errstr;
}

# fetch_page

sub fetch_page {
  my ( $self, $ua, $uri ) = @_;

  my $request;
# get HEADER
#  my $ua = $self->{ua};
  if ( $self->{expire}->{$uri} ) { # page was indexed
    my $h =  HTTP::Headers->new;
    $h->if_modified_since( $self->{expire}->{$uri} );

print "ifms:", $h->if_modified_since,"\n";

    $request  = HTTP::Request->new( GET => $uri, $h );
  } else {
    $request  = HTTP::Request->new( GET => $uri );
  }
  my $response = $ua->request($request);

  warn $uri,":",$response->as_string,"\n" if ( $self->{DEBUG} == 2 );
  my $reason = "";
  my $page = {};
  $page->{content_length} = $response->content_length;
  $page->{content_type} = $response->content_type;
  $page->{success} = $response->is_success;
  $page->{last_modified} = $response->last_modified;
  $page->{last_modified} ||= time(); # current time
  $page->{expires} = $response->expires;
  $page->{expires} ||= time(); # current time
  $page->{code} = $response->code;

  return $page if $response->content_type !~ /^text\// ||
                  !$response->is_success || 
                  $response->code == 304; # NOT MODIFIED

  $page->{uri} = ${ $response->request->url };
  $page->{base} = $response->base;
# $uri is probably changed due to redirection ! Need check against rules
  if ( $uri ne $page->{uri} ) {  
# add $page->{uri} to queue, delete $uri from queue and return (302) MOVED
    $self->delete_link($uri);
    $self->add_link( $page->{uri} );
  }
  $page->{content} = $response->content;
  $page->{title} = $response->title;
  $page->{md5} = md5_base64($page->{content});
# sometimes 'content_length' doesn't returned. Calculate it
  $page->{content_length} ||= length($page->{content});

  return $page;
}



1;

=head1 DESCRIPTION

 This package is designed to work with OpenFTS full text search engine.
 It inherits basic methods from Search::OpenFTS::Crawler.

 Overloaded methods - sql_patterns, insert, check_user_rules

=item B<WARNING:>

 Methods insert() and sql_patterns() should be consistent and reflect 
 user defined table scheme ( see init-crawler.pl )
 
 See crawler.pl for example of usage.

=cut

=head1 AUTHOR

 Oleg Bartunov (oleg@sai.msu.su)

=cut

=head1 SEE ALSO

 Search::OpenFTS::Crawler

=cut

